/*******************************************************************************
		Kepco TMA PC-27 Windows Driver - Kepco Inc. Rev. 4.0
********************************************************************************
Written by P.Blanco Kepco Inc. 1/97
Adapted from Dos Version. and modified for National Instruments.LabWindows/CVI
************************************************************************************
History:

REV. Date		By		Description                     
************************************************************************************
1.0  1/22/97   	P.B. 	Released 
2.0  2/28/97	P.B.	Reworked to make more modular to ease creating DLL
						for porting to labview. Split up Init and Reset to
						allow supplies to remain active if driver is stopped
						and restarted.
3.0  4/1/98		P.B.	Removed init routines from reset function, which was
						causeing bitbus to hang. Corrected problem in
						status reporting when run as a DLL.
3.1  1/28/99	P.B.	added model mst 6-24,50-3.5						
4.0  2/10/00    P.B.    added ReadStatus function, which basically just
						tells if the supply is present or not, and does not enable
						asyncronous status flags from the power module,as does the
						Status function. When working without an interrupt service 
						routine, it is better not to enable status reporting 
						( i.e. overload,voltage error etc). This is better handled
						by the user by just reading the volt,curr values and
						decideing if the supply is in the  right state.
						eliminated main, and replaced with KPTMAPC_panel
						provided means of selecting base address in function Init.
						added ID function,returns type,vmax,cmax
4.1	8/8/06	P.B.		changed volt_range & curr_range from WORD to float.
********************************************************************************/

#include <userint.h>
#include <utility.h>
#include <ansi_c.h>

#include "kptmapan.h"
#include "kptmapcw.h"

/** Power supply range table **/
/*  This table contains the max volt & curr for each Kepco Power supply model
	which works on the Kepco Control bus (Bitbus). The model ID is returned
	by the power supply during initialization.*/
	
struct
{
	float volt_range[64];
	float curr_range[64];
} range[4] =
{
	/**MST volt ranges **/
	6,15,25,36,55,75,100,150,
	325,15,25,36,15,40,40,65,
	6,50,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,

	/** MST current range **/
	20,12,8,5,3.5,2.5,2,1.2,
	0.6,10,6,4,2,5,10,5,
	24,3.5,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,

	/** BOP voltage range **/
	50,100,20,36,50,72,100,20,
	36,50,72,100,200,200,20,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,

	/** BOP current range **/
	2,1,10,6,4,3,2,20,
	12,8,6,4,0.5,1,5,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,

	/**MAT volt ranges **/
	6,15,25,36,55,75,100,150,
	6,15,25,36,55,75,100,150,
	6,15,25,36,55,75,100,150,
	8,8,30,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,

	/** MAT current range **/
	32,20,14,10,7,5,3.6,2.4,
	64,40,28,20,14,10,7.2,4.8,
	100,60,42,30,20,15,10,7,
	75,28,35,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,

	/**HSP volt ranges **/
	28,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,

	/** HSP current range **/
	36,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
};

/* data for each power supply */
typedef struct
{
	WORD last_status;	/** The last status byte received **/
	BYTE on_line;		/** The online condition **/
	BYTE identifier;	/** The identity code  **/
	BYTE type;			/** The model type */ 
	BYTE output;		/** Programmed output state **/
	BYTE mode;			/** Programmed mode volt/curr    **/
	BYTE polarity;  	/** Programmed polarity plus/min **/
	BYTE bb_status;     /** 0 if good, else 0x91 **/
	WORD volt_actual;	/** The actual voltage 0 to 0fff **/
	WORD curr_actual;	/** The actual current 0 to 0fff **/
	float volt_range;	/** The max value for voltage **/
	float curr_range;	/** The max value for current **/
} PowerSupply;

PowerSupply PS[32];

char model[4][4]=
{ "MST","BOP","MAT","HSP" };

int kpanel;												/* handle for main panel	*/
int tpanel,timer;										/* handles for timer panel	*/
int node;												/* Bitbus node #			*/
int sbx_base;											/* Bitbus base address		*/
int init_complete;										/* Initialize done flag		*/
int  bb_in_reset;										/* Reset in progress flag	*/
char psmodel[40];										/* Power Supply model		*/
int Max_Chan=32;										/* highest chan. #			*/ 	
BYTE id_flag,id[32];									/* Bitbus identity flags	*/
BYTE status_flag,status[32];							/* Bitbus status flags		*/
BYTE read_flag,read[32];								/* Bitbus measurement ready */
BYTE bbinpq[20];										/* Bitbus input buffer		*/
BYTE pending[32];										/* Bitbus command pending   */
BYTE bbqindex = 0;										/* Bitbus message index		*/
BYTE bbcommandheader[] = {12,0x40,0,1,0,0,0,0,0,0};		/* Bitbus command			*/
BYTE bbresetheader[] = {8,0x40,0,0,0,0};				/* Bitbus reset        		*/
BYTE ret_message[] = {1,0,1,1,1,0,1,0,0,0,0};			/* weather cmnd has reply   */
BYTE status_latch;
BYTE last_node_checked;									/* node index for poll_task	*/
BYTE DLL_MODE = 1;										/* 1 when DLL,0 when EXE	*/
int holdoff=0;
/************************************************************************

	FUNCTION : int waitresetresponse( BYTE node );

	INPUTS : BYTE node

	OUTPUTS : 0 = ok  -1 = error

	DESCRIPTION : wait 100 milliseconds to make sure that the board
                  was reset properly and not sending any response.  The
		  proper response to a reset command is no message.


*************************************************************************/
int waitresetresponse(BYTE node)
{
	WORD count;

	if (node > Max_Chan)
	{
		node = 0;
	}

	/** Check for a response - up to 200 milliseconds **/
	for(count=0;count<20;count++)
	{

		if (PS[node].bb_status != 0)
		{
			return (-1);
		}
		Delay(0.01);
	}

	/** No response, return 0 **/
	return (0);

}

/************************************************************************

	FUNCTION : int resetboard(BYTE node)

	INPUTS : BYTE node - The bitbus node to be reset.

	OUTPUTS : 0 = ok  -1 = error

	DESCRIPTION : Send the reset message to the node requested.


*************************************************************************/
int resetboard(BYTE node)
{
	BYTE ramresetmsg[6];

	/** Move command/bytes **/
	memcpy(ramresetmsg,bbresetheader,6);
	ramresetmsg[2]=node;

	/** Transmit the reset command **/
	if (bitbus_out(ramresetmsg)== -1);
		//return(-1);					

	/** Wait for a response **/
	return (waitresetresponse(node));
}

/************************************************************************

	FUNCTION : int reset_344(void);

	INPUTS : bb_interrupt_number

	OUTPUTS : status : 0 OK; 1 error.

	DESCRIPTION : Flush all outstanding messages in the fifo, then
				  send the reset command to the 344 board.
				  Then setup receive bitbus interrupt vector.


*************************************************************************/
BYTE reset344(void)
{
	WORD count;


	count = 0;

	/**  flush all messages from the 344 fifo **/
	while (((inp(STATPORT) & (RFNE | RCMD)) != (RFNE | RCMD)) && (count++<50000))
	{
		/** If RCMD bit is low, then input from COMMPORT **/
		if ((inp(STATPORT) & RCMD) == 0)
			inp(COMMPORT);

		/** Else this is data **/
		else
			inp(DATAPORT);

		Delay(0.01);

	}
	/** Check for invalid board data **/
	if (inp(STATPORT)==0xff)
		return(1);

	if ((inp(STATPORT) & (RFNE | RCMD)) != (RFNE | RCMD))
		return(1);

	/** Reset the 344 board **/
	resetboard(0xFF);

	/** Wait 1 seconds while the 344 resets **/
	Delay(1);

	return (0);

}

int bitbus_out(BYTE *bbsend) /* sends a message out to a power supply */
{
	int  i,len;
	long cnt;
	len = (int)bbsend[0]-2;
	cnt = 0;
	
	SuspendTimerCallbacks ();
	
	while ((inp(STATPORT) & TFNF) != 0)
	{
		if(++cnt==MAXCOUNT)
			break;
			//return(-1);
	}
	if(cnt==MAXCOUNT) MessagePopup ("Kepco TMA PC-27", "Communication Error");
	for(i=0;i<len;i++)
	{
     	cnt = 0;
     	while ((inp(STATPORT) & TFNF) != 0)
		{
			if(++cnt==MAXCOUNT)
				break;
				//return(-1);
		}
		outp(DATAPORT,bbsend[i] & 0xFF);
	}
    cnt = 0;
    while ((inp(STATPORT) & TFNF) != 0)
	{
		if(++cnt==MAXCOUNT)
			break;
			//return(-1);
	}
    outp(COMMPORT,0);
	
	if(bb_in_reset==0)
	ResumeTimerCallbacks ();

	if(cnt==MAXCOUNT) return(-1);
	return(0);
}

int bb_xmit(BYTE command, BYTE node, WORD param1, WORD param2)
{
	int i,status;
	
	struct
	{
		BYTE length;
		BYTE route;
		BYTE node;
		BYTE tasks;
		BYTE message[6];
	} bb_xmt;
	
	char dbstring[40];
	
	if(node<32)
	if(PS[node].on_line!=2 && bb_in_reset==0) return 0;

	if((command==BB_RELAY)&&( (PS[node].identifier & 0x40) ==0x40))
		return 0;


	if (command < 10 && ret_message[command] > 0)
	{
        i=0;
		while (pending[node] && ++i< 30) Delay(.01);
		if(i>=30) Beep();
		pending[node]=1;
	}
	if(command < 10 || command > 13)
	{
		memcpy(&bb_xmt,&bbcommandheader,sizeof(bbcommandheader));
		bb_xmt.message[1] = command;
		bb_xmt.message[2] = param1 >> 8;
		bb_xmt.message[3] = (BYTE)param1;
		bb_xmt.message[4] = param2 >> 8;
		bb_xmt.message[5] = (BYTE)param2;
	}
	else if(command == 10) memcpy(&bb_xmt,&bbresetheader,sizeof(bbresetheader));
	
	bb_xmt.node = node;
	sprintf(dbstring,"%d,%d",node,command);
	status= bitbus_out((char *)&bb_xmt);
	if( status== -1){
		
		i=0;
		while(i++<10)
		Beep();
		//reset344();
	}
	return status;
}

void poll_task(void)
{
	/* This routine is called by a timer task to periodically poll each
		currently active power supply node on the Kepco control bus (Bitbus).
		If this is not done the power supply will timeout after about 60 sec.
		and will reset itself.1 supply is polled each time.*/

	static int pollcount,i;
	
	if (init_complete && pollcount++ >100) /* poll about 1/sec. timertick=10msec*/
	{
		pollcount=0;
		if (bb_in_reset == 0)
		{
			last_node_checked = (last_node_checked+1) % (Max_Chan-1);
			if (PS[last_node_checked+1].on_line == 2)
				bb_xmit(BB_ATTENDANCE,last_node_checked+1,0,0);
		}
	}
}

/* 	This routine is called periodically by a timer task to look for data comming
	back from the power supplies. */

void bb_rcv_task(void) 
{
	BYTE sbxstatus;
	BYTE response;
	BYTE task;
	BYTE node;
	BYTE model,type;
	int  command,i;
	long wait;

	
	while(1)
	{
		wait = 0;
		do				   
		{
			if (++wait >300) return;
			sbxstatus=(inp(STATPORT));
		} while ((sbxstatus & RFNE) != 0);
		if((sbxstatus & RCMD) == 0)
		{
			inp(COMMPORT);
			break;
		}
		else
		{
			bbinpq[bbqindex++] = inp(DATAPORT);
		}
		if(bbqindex>=MAXSIZE)
		{
			inp(COMMPORT);
			break;
		}
	}
	node = bbinpq[2];
	task = bbinpq[3];
	command = bbinpq[4];
	response = bbinpq[5];
	bbqindex = 0;

	if (response==0xFF)
		pending[node]=0;
	if (task==0)
	{
		if(command==0x91)
		{
			if (node > MAXNODE)
				node = 0;
			PS[node].on_line=0;
			PS[node].bb_status = 0x91;
		}
	}
	else if (command>0)
	{
		if(node>0 && node<Max_Chan)
		{
			pending[node] = 0;
			PS[node].on_line = 0;
		}
	}
	else
	{
		if(response==BB_IDENTIFIER)
		{
			PS[node].identifier = bbinpq[6];
			PS[node].on_line = 2;
			type  = ( (bbinpq[6] & 0x80) / 0x80)*2 + ( (bbinpq[6] & 0x40) / 0x40);
			model = bbinpq[6] & 0x1F;


			/** Set up the PS voltage and current range **/
			PS[node].volt_range = range[type].volt_range[model];
			PS[node].curr_range = range[type].curr_range[model];
			PS[node].type = type;
			id_flag=1;
			id[node]=1;				


			/** Send a status request message to this power supply **/
			//bb_xmit(BB_STATUS_PS,node,0,0);
		}

		if(response==BB_STATUS_PS)
		{
			PS[node].last_status = (bbinpq[7]*256)+ bbinpq[6] ;
            status_flag=1;
            status[node]=1;
			/*
			if(DLL_MODE) // v3.0
				bb_xmit(BB_STATUS_PS,node,0,0);
			*/				
				
		}
		else if(response == BB_READ)
		{
			PS[node].volt_actual=(WORD)(bbinpq[6])*256+(WORD)(bbinpq[7]);
			PS[node].curr_actual=(WORD)(bbinpq[8])*256+(WORD)(bbinpq[9]);
			read_flag=1;
        	read[node]=1;
		}
	}
}
int bb_init()
{
	int i,status;

	SuspendTimerCallbacks ();
	bb_in_reset = 1;

	for(i=1;i<Max_Chan;i++){
		if(PS[i].on_line != 2 )
		{
			PS[i].identifier = 0xFF;
			PS[i].on_line = 3;
			pending[i]=0;
			if(bb_xmit(BB_ATTENDANCE,i,0,0)==-1) return -1;//changed from BB_NODE_RESET v1.1
			Delay(.03);
			bb_rcv_task();
		}
	}
	if(init_complete == 0)
	{
		bb_xmit(BB_NODE_RESET,255,0,0);
		Delay(1);
	}
	for(i=1;i<Max_Chan;i++)
	{
		if(PS[i].on_line == 3)
		{
			bb_xmit(BB_IDENTIFIER,i,0,0);
			Delay(.03);
			bb_rcv_task();
		}
	}

	bb_in_reset = 0;
	init_complete = 1;
	ResumeTimerCallbacks ();
	return 0;
}
void bb_reset(int mode)
{
	int i;

	SuspendTimerCallbacks ();
	bb_in_reset = 1;

	for(i=1;i<Max_Chan;i++)
	{
		if(PS[i].on_line == 2)
		{
			bb_xmit(BB_RESET_PS,i,0,0);
		}
	}


	for (i=1;i<Max_Chan;i++){
		PS[node].output=0;
		PS[node].polarity=0;
		PS[node].mode=0;
	}

	Delay(2);
	bb_in_reset = 0;
	ResumeTimerCallbacks ();

}
void KPTMAPC_TimerTask(void){
	
	bb_rcv_task();
	poll_task();
	if(holdoff)holdoff--;
}
int kp_timer (int panel, int control, int event,
		void *callbackData, int eventData1, int eventData2)
{
	static int i,readcount;
	float volt,curr;
	switch (event) {
		case EVENT_TIMER_TICK:
			KPTMAPC_TimerTask();
			if(PS[node].on_line==2)
			SetCtrlVal (kpanel, kptmapan_online_led,1);
			else
			SetCtrlVal (kpanel, kptmapan_online_led,0);
			if(read[node]!=1 && ++readcount>10 && !holdoff){
			readcount=0;
			bb_xmit(7,node,0,0);
			}
			else
			if(read[node]){
			read[node]=0;
			volt=(float)(PS[node].volt_actual&0xfff)/4095*PS[node].volt_range;
			curr=(float)(PS[node].curr_actual&0xfff)/4095*PS[node].curr_range;
			if(PS[node].volt_actual&0x1000) volt*=-1;
			if(PS[node].curr_actual&0x1000) curr*=-1;
			SetCtrlVal (kpanel, kptmapan_Voltage_read,volt);
			SetCtrlVal (kpanel, kptmapan_Current_read,curr);
			}	
			SetCtrlVal (kpanel, kptmapan_Output_led,PS[node].output);
			SetCtrlVal (kpanel, kptmapan_Polarity_led,PS[node].polarity);
			SetCtrlVal (kpanel, kptmapan_Voltage_led,(!PS[node].mode));
			SetCtrlVal (kpanel, kptmapan_Current_led,PS[node].mode);
			if(status_flag){
       		status_flag=0;
        	for(i=0;i<Max_Chan;i++)
          		if(status[i]){
       	  			status[i]=0;
					//bb_xmit(BB_STATUS_PS,i,0,0);
				}
			}
    		if(id_flag){
   				id_flag=0;
				if(id[node]){
				sprintf(psmodel,"%s %4.1f-%4.1f",model[PS[node].type],
				(float)PS[node].volt_range,(float)PS[node].curr_range);
				SetCtrlVal (kpanel,kptmapan_Model,psmodel);
				}
			}
		    for(i=0;i<Max_Chan;i++)
       		if(id[i]){
   	  			id[i]=0;
	   	  	}
 			break;
	}
	return 0;
}

void init_timer(void)
{

	tpanel = NewPanel (0, "bb_timer",0,0 ,0 ,0 );
	timer = NewCtrl (tpanel, CTRL_TIMER, "", 0, 0);
	SetCtrlAttribute (tpanel, timer, ATTR_INTERVAL,.01 );
	
	InstallCtrlCallback (tpanel, timer, kp_timer, 0);

}

int KPTMAPC_Init(int addr)
{
	sbx_base = addr;
	return bb_init();
}
void KPTMAPC_ID(int node,int *type,float *volt,float *curr)
{
	if(PS[node].on_line == 2){
		*volt=(float)PS[node].volt_range;
		*curr=(float)PS[node].curr_range;
		*type=PS[node].type;
	}
	else{
		*volt=0;
		*curr=0;
		*type=0;
	}
	
}
void KPTMAPC_Reset(int node)
{
	if(!node)
	bb_reset(0);
	else
	if(PS[node].on_line == 2)
	{
		bb_xmit(BB_RESET_PS,node,0,0);
	}
	
}

void KPTMAPC_Set(int node,int mode,int pol,float volt,float curr)
{
	int vout,cout;
	float vmax,cmax;
	
	vmax=PS[node].volt_range;
	cmax=PS[node].curr_range;
	vout=abs(volt/vmax*4095);
	cout=abs(curr/cmax*4095);
	if(pol){vout|=0x1000;cout|=0x1000;}
	
	if(mode)
	bb_xmit(4,node,cout,vout);
	else
	bb_xmit(3,node,vout,cout);
}

void KPTMAPC_Output(int node,int mode)
{
	int cmnd;
	if(mode) cmnd=4095; else cmnd=0;
	bb_xmit(6,node,cmnd,cmnd);
}

void KPTMAPC_Read(int node,float *volt,float *curr)
{
	int i;
	read_flag=0;i=0;
	if(PS[node].on_line!=2){
		*volt=0;
		*curr=0;
		return;
	}
	bb_xmit(7,node,0,0);
    while(!read_flag && i++<1000)
        bb_rcv_task();
    
    if(read[node]){
       	read[node]=0;
		*volt=(float)(PS[node].volt_actual&0xfff)/4095*PS[node].volt_range;
		*curr=(float)(PS[node].curr_actual&0xfff)/4095*PS[node].curr_range;
		if(PS[node].volt_actual&0x1000) *volt*=-1;
		if(PS[node].curr_actual&0x1000) *curr*=-1;
    }
}
void KPTMAPC_ReadStatus(int node,WORD *status)
{
	if(PS[node].on_line==2) // just make sure its online
		*status=PS[node].last_status;
	else
		*status=2048;
}
void KPTMAPC_Status(int node,WORD *status)
{
	if(PS[node].on_line==2){
		bb_xmit(BB_STATUS_PS,node,0,0);
		Delay(.03);
		bb_rcv_task();
		*status=PS[node].last_status;
	}
	else
	*status=2048;
}

void KPTMAPC_Close(void)
{
	int i;
	SuspendTimerCallbacks ();
	bb_in_reset=1;
	for(i=1;i<Max_Chan;i++)
		if(PS[i].on_line==2)
			bb_xmit(9,i,0,0); /* Go To Local (prevents timeout reset) */
	DiscardPanel (tpanel);
}
int kp_channel (int panel, int control, int event,
		void *callbackData, int eventData1, int eventData2)
{
	switch (event) {
		case EVENT_COMMIT:
			GetCtrlVal (kpanel, kptmapan_Channel, &node);
			if(PS[node].on_line==2)
			sprintf(psmodel,"%s %4.1f-%4.1f",model[PS[node].type],
			(float)PS[node].volt_range,(float)PS[node].curr_range);
			else
			strcpy(psmodel,"No Device");
			SetCtrlVal (kpanel,kptmapan_Model,psmodel);
			
			break;
	}
	return 0;
}

int kp_outp (int panel, int control, int event,
		void *callbackData, int eventData1, int eventData2)
{
	int mode;
	
	switch (event) {
		case EVENT_COMMIT:
			mode=PS[node].output;
			if(mode==1)
				mode=0;
			else
				mode=1;
				
			KPTMAPC_Output(node,mode);
			PS[node].output=mode;
			break;
	}
	return 0;
}

int kp_exec (int panel, int control, int event,
		void *callbackData, int eventData1, int eventData2)
{
	float volt,curr;
	int mode=PS[node].mode;
																	    
	switch (event) {
		case EVENT_COMMIT:
			GetCtrlVal (kpanel, kptmapan_Voltage, &volt);
			GetCtrlVal (kpanel, kptmapan_Current, &curr);
			KPTMAPC_Set(node,mode,PS[node].polarity,volt,curr);
			//holdoff=200;
			break;
	}
	return 0;
}

int kp_mode (int panel, int control, int event,
		void *callbackData, int eventData1, int eventData2)
{
	float volt,curr;
	int mode;
	
	switch (event) {
		case EVENT_COMMIT:
			if(PS[node].mode==1)
				PS[node].mode=0;
			else
				PS[node].mode=1;
			GetCtrlVal (kpanel, kptmapan_Voltage, &volt);
			GetCtrlVal (kpanel, kptmapan_Current, &curr);
			KPTMAPC_Set(node,PS[node].mode,PS[node].polarity,volt,curr);
			break;
	}
	return 0;
}

int kp_polarity (int panel, int control, int event,
		void *callbackData, int eventData1, int eventData2)
{
	float volt,curr;
	int mode;
	switch (event) {
		case EVENT_COMMIT:
			if(PS[node].polarity==1)
				PS[node].polarity=0;
			else
				PS[node].polarity=1;
			mode=PS[node].mode;
			GetCtrlVal (kpanel, kptmapan_Voltage, &volt);
			GetCtrlVal (kpanel, kptmapan_Current, &curr);
			KPTMAPC_Set(node,mode,PS[node].polarity,volt,curr);
			break;
	}
	return 0;
}

int kp_reset (int panel, int control, int event,
		void *callbackData, int eventData1, int eventData2)
{
	switch (event) {
		case EVENT_COMMIT:
			bb_reset(0);
			break;
	}
	return 0;
}

int kp_init (int panel, int control, int event,
		void *callbackData, int eventData1, int eventData2)
{
	switch (event) {
		case EVENT_COMMIT:
			bb_init();
			break;
	}
	return 0;
}

int quit (int panel, int control, int event,
		void *callbackData, int eventData1, int eventData2)
{
	int i;
	switch (event) {
		case EVENT_COMMIT:
			SuspendTimerCallbacks ();
			bb_in_reset=1;
			for(i=1;i<Max_Chan;i++)
				if(PS[i].on_line==2)
					bb_xmit(9,i,0,0); /* Go To Local (prevents timeout reset) */
			QuitUserInterface (0);
			break;
	}
	return 0;
}

void InstallCallbacks()
{
	InstallCtrlCallback (kpanel, kptmapan_Stop, quit, 0);
	InstallCtrlCallback (kpanel, kptmapan_Init, kp_init, 0);
	InstallCtrlCallback (kpanel, kptmapan_Execute, kp_exec, 0);
	InstallCtrlCallback (kpanel, kptmapan_Reset, kp_reset, 0);
	InstallCtrlCallback (kpanel, kptmapan_Output, kp_outp, 0);
	InstallCtrlCallback (kpanel, kptmapan_Polarity, kp_polarity, 0);
	InstallCtrlCallback (kpanel, kptmapan_Mode, kp_mode, 0);
	InstallCtrlCallback (kpanel, kptmapan_Channel, kp_channel, 0);
}
void KPTMAPC_panel(int addr)
{
	DLL_MODE=0;
	KPTMAPC_Init(addr);
	init_timer();
	kpanel = LoadPanel (0, "kptmapan.uir", kptmapan);
	InstallCallbacks();
	GetCtrlVal (kpanel, kptmapan_Channel, &node);
	DisplayPanel (kpanel);
	RunUserInterface ();
}

void Wait(double seconds){
	Delay(seconds);
}
